#!/usr/bin/env python3
"""
Network Load Testing Tool
Parameters:
    - IP: Target IP address
    - port: Target port number
Additional parameters in comments:
    - protocol: TCP or UDP
    - duration: Test duration in seconds
    - threads: Number of concurrent connections
"""
import socket
import threading
import time
import argparse
import logging
from datetime import datetime
import json
from pathlib import Path
import queue
import statistics

class NetworkLoadTester:
    def __init__(self, target_ip, port, protocol='tcp', duration=60, threads=10,
                packet_size=1024, rate_limit=None):
        self.target_ip = target_ip
        self.port = port
        self.protocol = protocol.lower()
        self.duration = duration
        self.threads = threads
        self.packet_size = packet_size
        self.rate_limit = rate_limit
        
        self.stop_event = threading.Event()
        self.results_queue = queue.Queue()
        self.error_queue = queue.Queue()
        self.setup_logging()
        
    def setup_logging(self):
        """Set up logging configuration"""
        log_dir = Path("loadtest_logs")
        log_dir.mkdir(exist_ok=True)
        
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(log_dir / "loadtest.log"),
                logging.StreamHandler()
            ]
        )
        
    def create_socket(self):
        """Create appropriate socket based on protocol"""
        if self.protocol == 'tcp':
            return socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            return socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            
    def generate_payload(self):
        """Generate test payload"""
        return b'X' * self.packet_size
        
    def calculate_metrics(self, results):
        """Calculate performance metrics"""
        if not results:
            return None
            
        latencies = [r['latency'] for r in results if r['success']]
        if not latencies:
            return None
            
        return {
            'min_latency': min(latencies),
            'max_latency': max(latencies),
            'avg_latency': statistics.mean(latencies),
            'median_latency': statistics.median(latencies),
            'stddev_latency': statistics.stdev(latencies) if len(latencies) > 1 else 0,
            'successful_requests': len(latencies),
            'failed_requests': len(results) - len(latencies),
            'success_rate': len(latencies) / len(results) * 100
        }
        
    def worker_tcp(self, worker_id):
        """TCP connection worker"""
        sock = self.create_socket()
        payload = self.generate_payload()
        
        while not self.stop_event.is_set():
            try:
                start_time = time.time()
                
                # Connect
                sock.connect((self.target_ip, self.port))
                
                # Send data
                sock.send(payload)
                
                # Receive response
                response = sock.recv(self.packet_size)
                
                latency = time.time() - start_time
                
                self.results_queue.put({
                    'worker_id': worker_id,
                    'timestamp': time.time(),
                    'latency': latency,
                    'bytes_sent': len(payload),
                    'bytes_received': len(response),
                    'success': True
                })
                
                # Rate limiting
                if self.rate_limit:
                    time.sleep(1 / self.rate_limit)
                    
                sock.close()
                sock = self.create_socket()
                
            except Exception as e:
                self.error_queue.put({
                    'worker_id': worker_id,
                    'timestamp': time.time(),
                    'error': str(e)
                })
                self.results_queue.put({
                    'worker_id': worker_id,
                    'timestamp': time.time(),
                    'success': False
                })
                
                # Recreate socket on error
                sock.close()
                sock = self.create_socket()
                
    def worker_udp(self, worker_id):
        """UDP connection worker"""
        sock = self.create_socket()
        payload = self.generate_payload()
        
        while not self.stop_event.is_set():
            try:
                start_time = time.time()
                
                # Send data
                sock.sendto(payload, (self.target_ip, self.port))
                
                # Try to receive response (with timeout)
                sock.settimeout(2)
                response, _ = sock.recvfrom(self.packet_size)
                
                latency = time.time() - start_time
                
                self.results_queue.put({
                    'worker_id': worker_id,
                    'timestamp': time.time(),
                    'latency': latency,
                    'bytes_sent': len(payload),
                    'bytes_received': len(response),
                    'success': True
                })
                
                # Rate limiting
                if self.rate_limit:
                    time.sleep(1 / self.rate_limit)
                    
            except Exception as e:
                self.error_queue.put({
                    'worker_id': worker_id,
                    'timestamp': time.time(),
                    'error': str(e)
                })
                self.results_queue.put({
                    'worker_id': worker_id,
                    'timestamp': time.time(),
                    'success': False
                })
                
    def monitor_progress(self, start_time):
        """Monitor and display test progress"""
        while not self.stop_event.is_set():
            elapsed = time.time() - start_time
            remaining = max(0, self.duration - elapsed)
            
            # Get current metrics
            results = []
            while not self.results_queue.empty():
                results.append(self.results_queue.get())
                
            metrics = self.calculate_metrics(results)
            
            if metrics:
                print(f"\rProgress: {elapsed:.1f}/{self.duration}s "
                      f"Avg Latency: {metrics['avg_latency']*1000:.2f}ms "
                      f"Success Rate: {metrics['success_rate']:.1f}%", end="")
            else:
                print(f"\rProgress: {elapsed:.1f}/{self.duration}s", end="")
                
            time.sleep(1)
            
    def run_test(self):
        """Run the load test"""
        logging.info(f"Starting load test against {self.target_ip}:{self.port}")
        logging.info(f"Protocol: {self.protocol.upper()}")
        logging.info(f"Duration: {self.duration} seconds")
        logging.info(f"Threads: {self.threads}")
        logging.info(f"Packet Size: {self.packet_size} bytes")
        if self.rate_limit:
            logging.info(f"Rate Limit: {self.rate_limit} requests/second")
            
        start_time = time.time()
        workers = []
        
        # Start worker threads
        worker_func = self.worker_tcp if self.protocol == 'tcp' else self.worker_udp
        for i in range(self.threads):
            worker = threading.Thread(target=worker_func, args=(i,))
            worker.start()
            workers.append(worker)
            
        # Start progress monitor
        monitor = threading.Thread(target=self.monitor_progress, args=(start_time,))
        monitor.start()
        
        try:
            # Wait for duration
            time.sleep(self.duration)
            self.stop_event.set()
            
            # Wait for workers to finish
            for worker in workers:
                worker.join()
                
            monitor.join()
            
        except KeyboardInterrupt:
            logging.info("\nTest stopped by user")
            self.stop_event.set()
            for worker in workers:
                worker.join()
            monitor.join()
            
        # Collect all results
        results = []
        while not self.results_queue.empty():
            results.append(self.results_queue.get())
            
        errors = []
        while not self.error_queue.empty():
            errors.append(self.error_queue.get())
            
        # Calculate final metrics
        metrics = self.calculate_metrics(results)
        
        # Save results
        self.save_results(results, errors, metrics, time.time() - start_time)
        
    def save_results(self, results, errors, metrics, duration):
        """Save test results to file"""
        report = {
            'target': f"{self.target_ip}:{self.port}",
            'protocol': self.protocol,
            'timestamp': datetime.now().isoformat(),
            'duration': duration,
            'configuration': {
                'threads': self.threads,
                'packet_size': self.packet_size,
                'rate_limit': self.rate_limit
            },
            'metrics': metrics,
            'errors': errors[:100]  # Limit number of errors in report
        }
        
        report_file = Path("loadtest_logs") / f"report_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
        with open(report_file, 'w') as f:
            json.dump(report, f, indent=2)
            
        logging.info(f"\nTest completed. Report saved to {report_file}")
        if metrics:
            logging.info("\nFinal Metrics:")
            logging.info(f"Successful Requests: {metrics['successful_requests']}")
            logging.info(f"Failed Requests: {metrics['failed_requests']}")
            logging.info(f"Success Rate: {metrics['success_rate']:.1f}%")
            logging.info(f"Average Latency: {metrics['avg_latency']*1000:.2f}ms")
            logging.info(f"Min Latency: {metrics['min_latency']*1000:.2f}ms")
            logging.info(f"Max Latency: {metrics['max_latency']*1000:.2f}ms")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Network Load Testing Tool")
    parser.add_argument("ip", help="Target IP address")
    parser.add_argument("port", type=int, help="Target port number")
    parser.add_argument("--protocol", choices=['tcp', 'udp'], default='tcp',
                      help="Protocol to use (tcp or udp)")
    parser.add_argument("--duration", type=int, default=60,
                      help="Test duration in seconds")
    parser.add_argument("--threads", type=int, default=10,
                      help="Number of concurrent connections")
    parser.add_argument("--packet-size", type=int, default=1024,
                      help="Size of test packets in bytes")
    parser.add_argument("--rate-limit", type=float,
                      help="Rate limit in requests per second")
    args = parser.parse_args()
    
    tester = NetworkLoadTester(
        args.ip, args.port, args.protocol, args.duration,
        args.threads, args.packet_size, args.rate_limit
    )
    tester.run_test()
